/***************************************************************************/
/* initializeRenderingData.cpp                                             */
/* -----------------------                                                 */
/*                                                                         */
/* Initialization routines for data needed for all the rendering methods   */
/*    in this directory.                                                   */
/*                                                                         */
/* Chris Wyman (02/23/2008)                                                */
/***************************************************************************/


#include "sceneLoader.h"
#include "renderingData.h"
#include "renderingMethods.h"

extern Scene *scene;
RenderingData *data = 0;


void InitializeRenderingData( void )
{
	// Check to make sure we only initialize once.
	if (data) return;

	// Allocate global data structures
	data = new RenderingData();
	data->fbo    = new FrameBufferData();
	data->param  = new ParameterData();
	data->shader = new ShaderData();
	data->tex    = new TextureData();
	data->vbo    = new VertexBufferData();
	data->glID   = new OtherGLIDData();
	data->ui     = new UIData();

	// Allocate the temporary framebuffers used in our framework

	// The main window.  It seems sort of silly to have a "main window" buffer that is later 
	//    copied to the screen, but this gives us great flexibility when testing things out, so
	//    we find the extra overhead worth it.
	data->fbo->mainWin = new FrameBuffer( GL_TEXTURE_2D, scene->GetWidth(), scene->GetHeight(), -1, 
										  GL_RGBA, 1, 1, 0, "Main Render Window" );
	data->fbo->mainWin->CheckFramebufferStatus( 1 );

	// This buffer stores our shadow map that's computed ever frame.  Notice we use a pretty high
	//    resolution buffer to get shadows that are relatively alias free on surfaces in our renderings
	data->fbo->shadowMap = new FrameBuffer( GL_TEXTURE_2D, 2*scene->GetWidth(), 2*scene->GetHeight(), -1, 
											GL_LUMINANCE, 1, 1, 0, "Shadow Map" );
	data->fbo->shadowMap->CheckFramebufferStatus( 1 );

	// This buffer stores UI help information.  I've found that printing text to the screen using
	//   GLUT can be pretty slow, so I "precompute" the help screen into this buffer to keep 
	//   the application interactive while displaying our help screen
	data->fbo->helpScreen = new FrameBuffer( GL_TEXTURE_2D, 1024, 1024, -1, 
										  GL_RGBA, 1, 1, 0, "Help Screen" );
	data->fbo->helpScreen->CheckFramebufferStatus( 1 );

	// This buffer stores a rendering of the scene with no fog.  This is the first past, after which
	//    a scattering pass occurs, and the final pass combines this texture and the scattering texture 
	//    using a filter to upsample the scattering.
	data->fbo->volShad_firstPass = new FrameBuffer( GL_TEXTURE_2D, scene->GetWidth(), scene->GetHeight(), 
													-1, GL_RGBA16F_ARB, 1, 1, 0, "Volume Shadows First Pass" );
	data->fbo->volShad_firstPass->CheckFramebufferStatus( 1 );

	// This stores the per-frame reparameterized F texture.  Currently this is unused, since we
	//    use just the Hoffman and Preetham scattering model in this demo.  To use the Sun et al model instead,
	//                  see the comments in the basicVolumeShadows.cpp and volumetricShadows.cpp files as
	//                  well as the associated shaders: computeLowResScattering_PointLightFrontToBackVol.frag.glsl
	//                  and createLowResScatteringTextureUsingShadowVolumes.frag.glsl
	data->fbo->fog_perFrameFTex = new FrameBuffer( GL_TEXTURE_2D, 256, 256, 
												   -1, GL_RGBA16F_ARB, 1, 0, 0, "Per-Frame Airlight Texture" );
	data->fbo->fog_perFrameFTex->CheckFramebufferStatus( 1 );

	// This buffer stores information about the light frustum
	data->fbo->volShad_lightVol = new FrameBuffer( GL_TEXTURE_2D, 
		                                           data->fbo->fog_perFrameFTex->GetWidth(), 
												   data->fbo->fog_perFrameFTex->GetHeight(), 
												   -1, GL_RGBA16F_ARB, 1, 1, 0, "Light Volume Data" );
	data->fbo->volShad_lightVol->CheckFramebufferStatus( 1 );

	// This buffer stores the distances to the front & back shadow volumes
	data->fbo->volShad_shadowVolFrontBack = new FrameBuffer( GL_TEXTURE_2D_ARRAY_EXT, 
		                                           data->fbo->fog_perFrameFTex->GetWidth(), 
												   data->fbo->fog_perFrameFTex->GetHeight(), 
												   2, GL_LUMINANCE16F_ARB, 1, 1, 0, "Double-Sided Volume Depth Test FB" );
	data->fbo->volShad_shadowVolFrontBack->CheckFramebufferStatus( 1 );

	// This buffer stores the fog scattering contributions computed at a low resolution
	data->fbo->volShad_lowResFog = new FrameBuffer( GL_TEXTURE_2D, 256, 256,
												   -1, GL_RGBA16F_ARB, 2, 0, 0, "Low Resolution Fog Scattering" );
	data->fbo->volShad_lowResFog->CheckFramebufferStatus( 1 );


	// Allocate various parameter variable used in the code.  First we check to see
	//    if they were previously defined in the scene file.  Otherwise, we'll define 
	//    and allocate them here.
	data->param->usingShadows = scene->GetSceneBoolVar( "shadowsenabled" );
	if (!data->param->usingShadows) 
		data->param->usingShadows = new bool( true );
	data->ui->animation = scene->GetSceneBoolVar( "animation" );
	if (!data->ui->animation)
		data->ui->animation = new bool( true );
	data->param->shadowMapBias = scene->GetSceneFloatVar( "shadowmapbias" );
	if (!data->param->shadowMapBias)
		data->param->shadowMapBias = new float( -0.005 );
	data->param->lightFOV = scene->GetSceneFloatVar( "lightfov" );
	if (!data->param->lightFOV)
		data->param->lightFOV = new float( 70.0 );
	data->ui->displayID = scene->GetSceneIntVar( "displayid" );
	if (!data->ui->displayID ) data->ui->displayID  = new int( 0 );


	// Initialize and/or allocate various textures.  The spotlight textures are only loaded
	//    here if the scene file didn't explcitly specify which textures to use as spotlights.
	data->tex->spotlight = scene->GetNamedTexture( "spotlight" );
	if (!data->tex->spotlight)
	{
		char *file = scene->paths->GetTexturePath( "spot_white.ppm" );
		data->tex->spotlight = new GLTexture( file, 
		                                      TEXTURE_REPEAT_S | TEXTURE_REPEAT_T | TEXTURE_MIN_LINEAR_MIP_LINEAR );
		free( file );
	}
	data->tex->lowResSpot = scene->GetNamedTexture( "spotlightlowres" );
	if (!data->tex->lowResSpot)
	{
		char *file = scene->paths->GetTexturePath( "spot_white64.ppm" );
		data->tex->lowResSpot = new GLTexture( file, 
		                                      TEXTURE_REPEAT_S | TEXTURE_REPEAT_T | TEXTURE_MIN_LINEAR_MIP_LINEAR );
		free( file );
	}

	// Textures for "advertising" our affiliations in the help screen.
	char *file = scene->paths->GetTexturePath( "iowaLogo.ppm" );
	data->tex->iowaLogo = new GLTexture( file );
	free( file );
	file = scene->paths->GetTexturePath( "washColIcon.ppm" );
	data->tex->washColLogo = new GLTexture( file );
	free( file );


	// Initialize various other OpenGL identifier data
	glGenQueries( 1, &data->glID->primCountQuery  );
	glGenTextures( 1, &data->glID->airlightTexture );
	SetupAirlightModel_FTexture( data->glID->airlightTexture );

	// Initialize starting UI information
	data->ui->captureScreen = false;
	data->ui->displayHelp   = true;
	data->ui->frameSpeed = new FrameRate( 50 ); 
	data->ui->frameGrab  = new FrameGrab();     

	// Allocate and initialize various shaders

	// This shader reparameterizes the pretabulated F texture to make use when computer volume shadows faster
	//    Please note:  Due to simplifications of the code to make the demo clearer, this is no longer used,
	//                  (the Hoffman and Preetham scattering model is always used).  To use this model instead,
	//                  see the comments in the basicVolumeShadows.cpp and volumetricShadows.cpp files as
	//                  well as the associated shaders: computeLowResScattering_PointLightFrontToBackVol.frag.glsl
	//                  and createLowResScatteringTextureUsingShadowVolumes.frag.glsl
	scene->AddShader( data->shader->fog_CreatePerFrameAirlightTex =
                      new GLSLProgram( NULL, 
									   NULL, 
									   "createPerFrameAirlightTex.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->fog_CreatePerFrameAirlightTex->SetProgramDisables( GLSL_LIGHTING | GLSL_DEPTH_TEST );
	data->shader->fog_CreatePerFrameAirlightTex->SetupAutomaticBinding( "fTexRange", 4, data->param->fog_FTexRange );

	// This shader stores information about the light's view frustum to avoid ray-marching through
	//    regions of the scene that could not be lit.
	scene->AddShader( data->shader->volShad_storeLightVol =
		              new GLSLProgram( "drawLightVolumes.vert.glsl",  
									   NULL, 
									   "drawLightVolumes.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->volShad_storeLightVol->SetProgramDisables( GLSL_LIGHTING | GLSL_DEPTH_TEST | GLSL_CULL_FACE );
	data->shader->volShad_storeLightVol->SetProgramEnables( GLSL_BLEND );

	// The shader to create the low resolution fog image under a complex, textured light source
	scene->AddShader( data->shader->volShad_computeLowResScatteringWithShadVol =
		              new GLSLProgram( NULL,  
									   NULL, 
									   "createLowResScatteringTextureUsingShadowVolumes.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetProgramDisables( GLSL_LIGHTING | GLSL_DEPTH_TEST );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticBinding( "epS", 4, &data->param->shadowMatrixTrans[0] ); 
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticBinding( "epT", 4, &data->param->shadowMatrixTrans[4] ); 
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticBinding( "epR", 4, &data->param->shadowMatrixTrans[8] ); 
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticBinding( "epQ", 4, &data->param->shadowMatrixTrans[12] ); 
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticTextureBinding( "objColorAndDist", data->fbo->volShad_firstPass->GetColorTextureID(0), GL_TEXTURE0, GL_TEXTURE_2D );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticTextureBinding( "shadowMap", data->fbo->shadowMap->GetDepthTextureID(), GL_TEXTURE3, GL_TEXTURE_2D );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticTextureBinding( "newFogTex", data->fbo->fog_perFrameFTex->GetColorTextureID(0), GL_TEXTURE5, GL_TEXTURE_2D );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticTextureBinding( "lightTexColor", 
						( data->tex->lowResSpot ? data->tex->lowResSpot->TextureID() : data->tex->spotlight->TextureID() ), 
						GL_TEXTURE4, GL_TEXTURE_2D );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticTextureBinding( "shadowVolumeDists", data->fbo->volShad_shadowVolFrontBack->GetColorTextureID(0), GL_TEXTURE1, GL_TEXTURE_2D_ARRAY_EXT );
	data->shader->volShad_computeLowResScatteringWithShadVol->SetupAutomaticTextureBinding( "lightVolBounds", data->fbo->volShad_lightVol->GetColorTextureID(0), GL_TEXTURE2, GL_TEXTURE_2D );

    // The shader to create the low resolution fog image under a simple, point light source
	scene->AddShader( data->shader->volShad_computeBasicLowResScattering = 
                      new GLSLProgram( NULL,  
									   NULL, 
									   "computeLowResScattering_PointLightFrontToBackVol.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->volShad_computeBasicLowResScattering->SetProgramDisables( GLSL_LIGHTING | GLSL_DEPTH_TEST );
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticBinding( "epS", 4, &data->param->shadowMatrixTrans[0] ); 
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticBinding( "epT", 4, &data->param->shadowMatrixTrans[4] ); 
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticBinding( "epR", 4, &data->param->shadowMatrixTrans[8] ); 
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticBinding( "epQ", 4, &data->param->shadowMatrixTrans[12] ); 
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticTextureBinding( "objColorAndDist", data->fbo->volShad_firstPass->GetColorTextureID(0), GL_TEXTURE0, GL_TEXTURE_2D );
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticTextureBinding( "shadowMap", data->fbo->shadowMap->GetDepthTextureID(), GL_TEXTURE3, GL_TEXTURE_2D );
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticTextureBinding( "newFogTex", data->fbo->fog_perFrameFTex->GetColorTextureID(0), GL_TEXTURE5, GL_TEXTURE_2D );
	data->shader->volShad_computeBasicLowResScattering->SetupAutomaticTextureBinding( "shadowVolumeDists", data->fbo->volShad_shadowVolFrontBack->GetColorTextureID(0), GL_TEXTURE1, GL_TEXTURE_2D_ARRAY_EXT );

	// The shader to combine low res scattering with un-attenuated scene geometry using a bilateral-like filter
	scene->AddShader( data->shader->volShad_lowResPlusGeometry = 
					  new GLSLProgram( NULL,  
									   NULL, 
									   "addSceneAndLowResFog.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->volShad_lowResPlusGeometry->SetProgramDisables( GLSL_LIGHTING | GLSL_DEPTH_TEST );
	data->shader->volShad_lowResPlusGeometry->SetupAutomaticTextureBinding( "objColorAndDist", data->fbo->volShad_firstPass->GetColorTextureID(0), GL_TEXTURE6, GL_TEXTURE_2D );
	data->shader->volShad_lowResPlusGeometry->SetupAutomaticTextureBinding( "lightVolBounds" , data->fbo->volShad_lightVol->GetColorTextureID(0), GL_TEXTURE2, GL_TEXTURE_2D );
	data->shader->volShad_lowResPlusGeometry->SetupAutomaticTextureBinding( "shadowMap", data->fbo->shadowMap->GetDepthTextureID(), GL_TEXTURE3, GL_TEXTURE_2D );
	data->shader->volShad_lowResPlusGeometry->SetupAutomaticTextureBinding( "newFogTex", data->fbo->fog_perFrameFTex->GetColorTextureID(0), GL_TEXTURE5, GL_TEXTURE_2D );

	// This shader takes in edges from a polygonal model, determines which are silhouettes and extrudes 
	//     them away from the light to become the shadow volume
	scene->AddShader( data->shader->volShad_computeShadowVolFrontAndBack = 
		              new GLSLProgram( "findFrontAndBackVols.vert.glsl", 
									   "findFrontAndBackVols.geom.glsl", 
									   "findFrontAndBackVols.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->volShad_computeShadowVolFrontAndBack->GeometryShaderSettings( GL_LINES, 6, GL_TRIANGLE_STRIP );
	data->shader->volShad_computeShadowVolFrontAndBack->LinkProgram();
	data->shader->volShad_computeShadowVolFrontAndBack->SetProgramDisables( GLSL_CULL_FACE );
	data->shader->volShad_computeShadowVolFrontAndBack->SetProgramEnables( GLSL_DEPTH_TEST );

	// This shader takes in the polygonal model, and uses the fronts and backs to cap the near and
	//    far sides of the shadow volume created with the previously defined shader.
	scene->AddShader( data->shader->volShad_capShadowVol = 
		              new GLSLProgram( "renderGeomForShadVolCap.vert.glsl", 
									   "renderGeomForShadVolCap.geom.glsl", 
									   "renderGeomForShadVolCap.frag.glsl", 
									   true, scene->paths->GetShaderPathList() ) );
	data->shader->volShad_capShadowVol->GeometryShaderSettings( GL_TRIANGLES, 6, GL_TRIANGLE_STRIP );
	data->shader->volShad_capShadowVol->LinkProgram();
	data->shader->volShad_capShadowVol->SetProgramDisables( GLSL_CULL_FACE );
	data->shader->volShad_capShadowVol->SetProgramEnables( GLSL_DEPTH_TEST );

	// Setup the screen that displays help on key commands
	SetupHelpScreen( data->fbo->helpScreen );
}




