/***************************************************************************/
/* utilRoutines.cpp                                                        */
/* -----------------------                                                 */
/*                                                                         */
/* These are utility routines that may (or may not) be reused in multiple  */
/*    techniques in the various files in this directory.  However, these   */
/*    functions are not particularly relevant to the renderings, but       */
/*    instead encapsulate annoying OpenGL code that would otherwise make   */
/*    the rendering code difficult to read.                                */
/*                                                                         */
/* Chris Wyman (02/23/2008)                                                */
/***************************************************************************/

#include "sceneLoader.h"
#include "renderingData.h"

extern Scene *scene;
extern RenderingData *data;


// Displays a volume surrounding the light's view frustum.
void DrawLightVolume( FrameBuffer *frame, unsigned int lightNum=0 )
{
	GLLight *l = scene->GetLight( lightNum );
	float lightNear = 0.01; // l->GetLightNearPlane();
	float lightFar  = l->GetLightFarPlane();
	float tanFovy_2 = tan( l->GetLightFovy() * M_PI / 360.0f );

	frame->BindBuffer();
	frame->ClearBuffers();
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	gluPerspective( scene->GetCamera()->GetFovy(), frame->GetAspectRatio(), 0.01, 3 * scene->GetCamera()->GetFar() );
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	scene->LookAtMatrix();
	scene->LightLookAtInverseMatrix( lightNum );

	data->shader->volShad_storeLightVol->EnableShader();
	glBlendFunc( GL_ONE, GL_ONE );
	glBegin( GL_QUADS );
		glVertex3f( tanFovy_2*lightNear, tanFovy_2*lightNear, -lightNear );
		glVertex3f( -tanFovy_2*lightNear, tanFovy_2*lightNear, -lightNear );
		glVertex3f( -tanFovy_2*lightFar, tanFovy_2*lightFar, -lightFar );
		glVertex3f( tanFovy_2*lightFar, tanFovy_2*lightFar, -lightFar );

		glVertex3f( tanFovy_2*lightNear, -tanFovy_2*lightNear, -lightNear );
		glVertex3f( tanFovy_2*lightFar, -tanFovy_2*lightFar, -lightFar );
		glVertex3f( -tanFovy_2*lightFar, -tanFovy_2*lightFar, -lightFar );
		glVertex3f( -tanFovy_2*lightNear, -tanFovy_2*lightNear, -lightNear );

		glVertex3f( tanFovy_2*lightNear, -tanFovy_2*lightNear, -lightNear );
		glVertex3f( tanFovy_2*lightNear, tanFovy_2*lightNear, -lightNear );
		glVertex3f( tanFovy_2*lightFar, tanFovy_2*lightFar, -lightFar );
		glVertex3f( tanFovy_2*lightFar, -tanFovy_2*lightFar, -lightFar );

		glVertex3f( -tanFovy_2*lightNear, -tanFovy_2*lightNear, -lightNear );
		glVertex3f( -tanFovy_2*lightFar, -tanFovy_2*lightFar, -lightFar );
		glVertex3f( -tanFovy_2*lightFar, tanFovy_2*lightFar, -lightFar );
		glVertex3f( -tanFovy_2*lightNear, tanFovy_2*lightNear, -lightNear );
	
		glVertex3f( -tanFovy_2*lightNear, tanFovy_2*lightNear, -lightNear );
		glVertex3f( tanFovy_2*lightNear, tanFovy_2*lightNear, -lightNear );
		glVertex3f( tanFovy_2*lightNear, -tanFovy_2*lightNear, -lightNear );
		glVertex3f( -tanFovy_2*lightNear, -tanFovy_2*lightNear, -lightNear );

		glVertex3f( tanFovy_2*lightFar, tanFovy_2*lightFar, -lightFar );
		glVertex3f( -tanFovy_2*lightFar, tanFovy_2*lightFar, -lightFar );
		glVertex3f( -tanFovy_2*lightFar, -tanFovy_2*lightFar, -lightFar );
		glVertex3f( tanFovy_2*lightFar, -tanFovy_2*lightFar, -lightFar );
	glEnd( );
	data->shader->volShad_storeLightVol->DisableShader();

	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );

	frame->UnbindBuffer();
}

// Applies the reparameterization of the F texture given a light position (in eye space) and
//     a maximal distance (usually the far plane)
void CreatePerFrameAirlightTexture( FrameBuffer *frame, float *eyeSpaceLightPos, float maxDist )
{
	frame->BindBuffer();
	frame->ClearBuffers();
	glMatrixMode( GL_PROJECTION );
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0,1,0,1);
	glMatrixMode( GL_MODELVIEW );
	glPushMatrix();
	glLoadIdentity();
	data->shader->fog_CreatePerFrameAirlightTex->EnableShader();
	data->shader->fog_CreatePerFrameAirlightTex->BindAndEnableTexture( "fTex", data->glID->airlightTexture, 
																		GL_TEXTURE0, GL_TEXTURE_2D );
	data->shader->fog_CreatePerFrameAirlightTex->SetParameter( "distMax", maxDist );  // The far distance.  
	data->shader->fog_CreatePerFrameAirlightTex->SetParameter( "imageSize", frame->GetWidth(), frame->GetHeight() );
	data->shader->fog_CreatePerFrameAirlightTex->SetParameterv( "lightPos", 3, eyeSpaceLightPos ); 
	data->shader->fog_CreatePerFrameAirlightTex->SetParameter( "extinctionCoef", scene->GetFogExtinctionCoef() );
	data->shader->fog_CreatePerFrameAirlightTex->SetParameter( "lightIntensity", scene->GetLightIntensityModifier() );
	glBegin( GL_QUADS );
		glTexCoord2f(0,0);	glVertex2f(0,0);
		glTexCoord2f(1,0);	glVertex2f(1,0);
		glTexCoord2f(1,1);	glVertex2f(1,1);
		glTexCoord2f(0,1);	glVertex2f(0,1);
	glEnd();
	data->shader->fog_CreatePerFrameAirlightTex->DisableTexture( GL_TEXTURE0, GL_TEXTURE_2D );
	data->shader->fog_CreatePerFrameAirlightTex->DisableShader();
	glMatrixMode( GL_PROJECTION );
	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );
	glPopMatrix();
	frame->UnbindBuffer();
}


// This reads in the precomputed tabulated F texture.  This texture was downloaded
//    from the Sun et al., project page on the Columbia website.
void SetupAirlightModel_FTexture( GLuint airlightID )
{
	float *fData, *FTexRange = data->param->fog_FTexRange;
	int fWidth, fHeight, count=0;

	// Open the fog/scattering datafile
	FILE * fDataFile = scene->paths->OpenTexture( "F_512.csv", "r" );
	if (!fDataFile)
		FatalError( "Unable to file file 'F_512.csv' containing scattering data!" );
	
	// Read the header
	fscanf( fDataFile, "%d %d %f %f %f %f", &fWidth, &fHeight, &FTexRange[0], &FTexRange[1], &FTexRange[2], &FTexRange[3] );

	// Allocate the memory and read the data in
	fData = (float *)malloc( fWidth * fHeight * sizeof( float ) );
	for (int j=0; j<fHeight; j++)
	{
		for (int i=0; i<fWidth-1; i++)
			fscanf( fDataFile, "%f,", &fData[count++] );
		fscanf( fDataFile, "%f", &fData[count++] );
	}

	fclose( fDataFile );

	glBindTexture( GL_TEXTURE_2D, airlightID );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE32F_ARB, fWidth, fHeight, 0, GL_LUMINANCE, GL_FLOAT, fData );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );

	free( fData );
}


// Displays the already created help screen
void DisplayHelpScreen( FrameBuffer *helpFB )
{
	glPushAttrib( GL_ALL_ATTRIB_BITS );
	glEnable(GL_BLEND);
	glBlendFunc( GL_ONE, GL_ONE_MINUS_SRC_ALPHA );
	helpFB->DisplayAsFullScreenTexture( FBO_COLOR0 );
	glPopAttrib();
}

// Sets up a prerendered help screen (drawing text with GLUT seems to be quite slow).
void SetupHelpScreen( FrameBuffer *helpFB )
{
	helpFB->BindBuffer();
	glClearColor( 0, 0, 0, 0 );
	helpFB->ClearBuffers();
	

	glColor4f(0.2, 0.2, 0.2, 0.8);
		glMatrixMode( GL_PROJECTION );
		glPushMatrix();
		glLoadIdentity();
		gluOrtho2D(0,1,0,1);
		glMatrixMode( GL_MODELVIEW );
		glPushMatrix();
		glLoadIdentity();
		glBegin( GL_QUADS );
			glTexCoord2f(0,0);	glVertex2f(0.05,0.05);
			glTexCoord2f(1,0);	glVertex2f(0.95,0.05);
			glTexCoord2f(1,1);	glVertex2f(0.95,0.95);
			glTexCoord2f(0,1);	glVertex2f(0.05,0.95);
		glEnd();

		glEnable( GL_TEXTURE_2D );
		glBindTexture( GL_TEXTURE_2D, data->tex->washColLogo->TextureID() );
		glEnable( GL_BLEND );
		glBlendFuncSeparate( GL_ONE, GL_ONE, GL_ZERO, GL_ONE );
		glBegin( GL_QUADS );
			glTexCoord2f(0.01,0.01);	glVertex2f(0.8,0.07);
			glTexCoord2f(0.99,0.01);	glVertex2f(0.93,0.07);
			glTexCoord2f(0.99,0.99);	glVertex2f(0.93,0.2);
			glTexCoord2f(0.01,0.99);	glVertex2f(0.8,0.2);
		glEnd();
		glBindTexture( GL_TEXTURE_2D, data->tex->iowaLogo->TextureID() );
		glBegin( GL_QUADS );
			glTexCoord2f(0.01,0.01);	glVertex2f(0.67,0.07);
			glTexCoord2f(0.99,0.01);	glVertex2f(0.8,0.07);
			glTexCoord2f(0.99,0.99);	glVertex2f(0.8,0.2);
			glTexCoord2f(0.01,0.99);	glVertex2f(0.67,0.2);
		glEnd();
		glDisable( GL_BLEND );
		glDisable( GL_TEXTURE_2D );

		glMatrixMode( GL_PROJECTION );
		glPopMatrix();
		glMatrixMode( GL_MODELVIEW );
		glPopMatrix();
		
	glColor4f(1,1,0,1);
	    DisplayString( 75, 950, "User Interface Commands:", 1024, 1024 );
	glColor4f(1,1,1,1);
		DisplayString( 120, 920, "Toggle on/off this help screen", 1024, 1024 );
		DisplayString( 120, 900, "Toggle on/off dragon rotation", 1024, 1024 );
		DisplayString( 120, 880, "Capture screenshot", 1024, 1024 );
		DisplayString( 120, 860, "Reload shaders (Try reloading if your driver had errors during shader compilation)", 1024, 1024 );
		DisplayString( 120, 840, "Quit demonstration program", 1024, 1024 );

	glColor4f(0,1,1,1);
		DisplayString( 75, 920, "[ h ]", 1024, 1024 );
		DisplayString( 75, 900, "[ a ]", 1024, 1024 );
		DisplayString( 75, 880, "[ f12 ]", 1024, 1024 );
		DisplayString( 75, 860, "[ r ]", 1024, 1024 );
		DisplayString( 75, 840, "[ q ]", 1024, 1024 );
		DisplayString( 75, 800, "[ up-arrow ]", 1024, 1024 );
		DisplayString( 75, 780, "[ down-arrow ]", 1024, 1024 );
		DisplayString( 75, 760, "[ > ] or [ . ]", 1024, 1024 );
		DisplayString( 75, 740, "[ < ] or [ , ]", 1024, 1024 );
		DisplayString( 75, 700, "[ left-mouse ]", 1024, 1024 );
		DisplayString( 75, 680, "[ middle-mouse ]", 1024, 1024 );
		DisplayString( 75, 660, "[ right-mouse ]", 1024, 1024 );

	glColor4f(1,1,1,1);
		DisplayString( 180, 800, "Increase fog density", 1024, 1024 );
		DisplayString( 180, 780, "Decrease fog density", 1024, 1024 );

		DisplayString( 180, 760, "Increase light intensity", 1024, 1024 );
		DisplayString( 180, 740, "Decrease light intensity", 1024, 1024 );

		DisplayString( 180, 700, "Rotate eye point around dragon using trackball", 1024, 1024 );
		DisplayString( 180, 680, "Rotate dragon using trackball", 1024, 1024 );
		DisplayString( 180, 660, "Rotate light using trackball", 1024, 1024 );

		DisplayString( 75, 620, "Three rendering modes are enabled in this demo program:", 1024, 1024 );
		DisplayString( 100, 605, "1) Standard OpenGL Rendering (no fog)", 1024, 1024 );
		DisplayString( 100, 590, "2) Fog with colored spotlight [Default]", 1024, 1024 );
		DisplayString( 100, 575, "3) Fog with white point light", 1024, 1024 );
		DisplayString( 75, 555, "Switch between these modes using", 1024, 1024 );

	glColor4f(0,1,1,1);
		DisplayString( 275, 555, "[ NumPad / ] and [ NumPad * ]", 1024, 1024 );

	glColor4f(1,1,1,1);
		DisplayString( 75, 515,  "Please note:", 1024, 1024);
		DisplayString( 100, 500, "1) Fog density and light intensity controls work slightly differently in each mode, as we modify", 1024, 1024);
		DisplayString( 120, 485, "the values so our defaults give relatively consistant shadow contrast in all modes.", 1024, 1024);
		DisplayString( 100, 470, "2) This prototype does not correctly handle being inside an object's shadow volume.", 1024, 1024);
		DisplayString( 100, 455, "3) There are a few views which need additional sampling to fully eliminate aliasing.  You can", 1024, 1024);
		DisplayString( 120, 440, "change the sampling rates defined in the shader files and reload using [ r ].", 1024, 1024);

		glColor4f(1,1,0,1);
		DisplayString( 675, 220,  "For more information, visit:", 1024, 1024);
		DisplayString( 690, 205,  "http://www.cs.uiowa.edu/~cwyman/publications/", 1024, 1024);


	helpFB->UnbindBuffer();

}


