/******************************************************************/
/* Scene.h                                                        */
/* -----------------------                                        */
/*                                                                */
/* The file defines a scene class that encapsulates all the       */
/*     information necessary to render an image with a ray tracer */
/* Also note that this class includes the TraceRay() method,      */
/*     which actually traces a ray through the scene.             */
/* Scenes should be setup similar to the example(s) in the file   */
/*     SceneSetup.cpp.                                            */
/*                                                                */
/* Chris Wyman (10/26/2006)                                       */
/******************************************************************/

#ifndef SCENE_H
#define SCENE_H

#include "DataTypes/Color.h"
#include "DataTypes/MathDefs.h"
#include "DataTypes/Array1D.h"
#include "Objects/Group.h"
#include "Objects/Object.h"
#include "Scene/Camera.h"
#include "Scene/GLLight.h"
#include "Materials/Material.h"
#include "Utils/ProgramPathLists.h"
#include "Utils/Trackball.h"
#include "Utils/glslProgram.h"


class FrameBuffer;

class Scene {
private:
	Camera *camera;           // Scene camera.  There is only one (for now).
	Group *geometry;          // Scene geometry.  There is only one, as it should be a container object.
	Array1D<GLLight *> light; // Scene lights.  A list of all lights in the scene.

	// These are used for loading the scene into memory from a file
	Array1D<Material *   > fileMaterials;  // a list of materials, so they can be named and reused
	Array1D<GLTexture *  > fileTextures;   // a list of textures, so they can be reused
	Array1D<GLSLProgram *> fileShaders;    // a list of shaders, so they can be reloaded when needed

	// The following variables are declared in the scene file.  These variables have an associated
	//    data type and a character string name.  A search through the identifier list gives the
	//    array index of the identifier, which is the same index as the corresponding data in the
	//    data array.  
	// These are pretty ugly and need to be reworked (sometime) to make them more elegant and robust
	Array1D<char *>			floatVarNames;
	Array1D<float>			floatData;
	Array1D<char *>			intVarNames;
	Array1D<int>			intData;
	Array1D<char *>			boolVarNames;
	Array1D<bool>			boolData;
	bool sceneFileDataAccessed;

	// Used for loading objects.  Some may have names associated, in which case we may 
	//     need the names later in the file to reference them (e.g., for instances)
	Array1D<Object *> namedObjects;
	Array1D<char *>   objectNames;

	// Size of resulting image
	int screenWidth, screenHeight;

	// Describes verbosity of output.  Not all output to the console is
	//    surpressed with this flag.
	bool verbose;

	// Used if our scene has any spheres or cylinders
	GLUquadricObj *quadObj;

	// If we call CreateShadowMap, we'll store some temporary data here to allow usage
	//   of the shadow map by anyone with access to the scene data.  This could be 
	//   modified to encapsulate the data better, by making the scene object control
	//	 the shadow map.
	GLuint shadowMapTexID, causticMapTexID, causticDepthMapTexID;
	float shadMapTransMatrix[16];

	float *lightIntensityModifier, *extinctionCoef;
public:
	// Constructors & destructors
	Scene();										// Default constructor
	Scene( char *filename, bool verbose=false );    // Reads a scene from the file
	~Scene();

	// Get/set the camera
	inline Camera *GetCamera( void ) { return camera; }
	void SetCamera( Camera *cam );

	// Get the width and height of the image
	inline int GetWidth( void ) const			{ return screenWidth; }
	inline int GetHeight( void ) const			{ return screenHeight; }
	inline void SetWidth( int newWidth )		{ screenWidth = newWidth; }
	inline void SetHeight( int newHeight )		{ screenHeight = newHeight; }
	
	// Reload all the shaders
	bool ReloadShaders( void );

	// Get some default scene data (e.g., bgColor, rayDepth)
	//inline Color GetBackgroundColor( void ) const { return backgroundColor; }

	// Get/set the geometry.  Note, this is usually a group-type object
	inline Group *GetGeometry( void ) const                  { return geometry; }
	inline void SetGeometry( Group *obj )                    { geometry = obj; }

	// Deal with scene lights
	inline int GetNumLights( void ) const                    { return light.Size(); }
	inline void AddLight( GLLight *lght )                    { light.Add( lght ); }
	inline GLLight *GetLight( int i )                        { return light[i]; }
	inline float GetLightIntensityModifier( void )           { return *lightIntensityModifier; }
	inline void SetLightIntensityModifier( float newValue )  { *lightIntensityModifier = newValue; }

	inline float GetFogExtinctionCoef( void )	             { return *extinctionCoef; }
	inline void  SetFogExtinctionCoef( float newValue )      { *extinctionCoef = newValue; }

	// Setup trackballs for objects and lights.  These are called by object
	//    and/or light constructors when they find out the user wants to 
	//    associate a trackball with the object/light.  The scene takes care
	//    of setting things up correctly.
	// 'i' is a user spectified trackball ID.  Usually this works in conjunction
	//    with a scene-defined user interface that attaches some manipulation to
	//    this particular trackball ID.  A warning is printed if this is out of 
	//    the valid range.
	// The callee is responsible for creating the trackball and remembering a 
	//    pointer for any further use.
	void SetupLightTrackball( int i, Trackball *ball );
	void SetupObjectTrackball( int i, Trackball *ball );

	// Deal with scene textures & materials
	inline void AddTexture( GLTexture *tex )      { fileTextures.Add( tex ); }
	inline void AddMaterial( Material *mat )      { fileMaterials.Add( mat ); }
	inline void AddShader( GLSLProgram *shader )  { fileShaders.Add( shader ); }

	// Gets pointer to a variable defined in the scene file.
	float *GetSceneFloatVar( char *varName );
	int *GetSceneIntVar    ( char *varName );
	bool *GetSceneBoolVar  ( char *varName );


	// Setup light positions (there might be multiple of these, depending on
	//    if you want the Scene to setup the lights using the *current* modelview
	//    matrix or the *correct* modelview as the scene sees it)
	void SetupEnabledLightsWithCurrentModelview( void );

	// OpenGL scene setup
	inline void PerspectiveMatrix( void )       { gluPerspective( camera->GetFovy(), ((float)screenWidth)/screenHeight, camera->GetNear(), camera->GetFar() ); }
	inline void LookAtMatrix( void )            { camera->LookAtMatrix(); }
	inline void LookAtInverseMatrix( void )		{ camera->InverseLookAtMatrix(); }

	void LightPerspectiveMatrix( int i, float aspect );
	void LightPerspectiveInverseMatrix( int i, float aspect );
	void LightLookAtMatrix( int i );
	void LightLookAtInverseMatrix( int i );

	// Draw Geometry
	inline void Draw( unsigned int matlFlags=MATL_FLAGS_NONE, unsigned int optionFlags=OBJECT_OPTION_NONE, bool matlAlreadySpecified=false )							    
		{ geometry->Draw( this,matlFlags,optionFlags,matlAlreadySpecified ); }
	inline void DrawOnly( unsigned int propertyFlags, unsigned int matlFlags=MATL_FLAGS_NONE, unsigned int optionFlags=OBJECT_OPTION_NONE, bool matlAlreadySpecified=false ) 
		{ geometry->DrawOnly( this,propertyFlags,matlFlags,optionFlags,matlAlreadySpecified); }

	// Preprocess the scene
	void Preprocess( void );
	void SetupDefaultScenePaths( void );

	// Functions that are really only valid during scene preprocessing. 
	//    Calling at other times may return NULL values, invalid values, or
	//    otherwise corrupted data (i.e., undefined behavior)
	inline GLUquadricObj *GetQuadric( void )   { return quadObj; }

	// Process the scene between frames
	inline void PerFrameUpdate( float currentTime ) { if (geometry->NeedPerFrameUpdates()) geometry->Update( currentTime ); }

	// Used for loading the scene...  You may need to add more of these, or you 
	//    may be able to handle them using class constructors.
	Material *ExistingMaterialFromFile( char *name );
	GLTexture *ExistingTextureFromFile( char *filename );
	GLTexture *GetNamedTexture( char *name );
	Object *ExistingObjectFromFile( char *name );
	GLLight *ExistingLightFromFile( char *name );
	Material *LoadMaterial( char *typeString, FILE *file );
	Object *LoadObject( char *typeString, FILE *file );
	Camera *LoadCamera( char *typeString, FILE *file );
	Material *GetDefaultMaterial( void ) { return fileMaterials[0]; }

	// Some unhandled keywords are bracketed by an end.  They cannot simply
	//    be ignored -- all lines until an "end" must be ignored.  This 
	//    function ignores all such lines and prints a warning message.
	void UnhandledKeyword( FILE *f, char *typeString=0, char *keyword=0 );

	// To truly deal with unhandled keywords, we must deal with unhandled
	//    keywords with nested keyword blocks (e.g., a material defined 
	//    inside an object).  In these cases, searching for the first "end"
	//    doesn't work.  This checks if the keyword opens a block, and if
	//    so UnhandledKeyword() can be called recursively.
	bool IsBlockKeyword( char *keyword );


	// Rendering functions
	void CreateShadowMap( FrameBuffer *shadMapBuf, float *shadMapMatrixTranspose, int lightNum, float shadowMapBias=0 );

	// Get/Set data for various rendered buffers.  Note CreateShadowMap() sets the appropriate
	//   values for the shadow map data...
	inline GLuint GetShadowMapID( void ) const			{ return shadowMapTexID; }
	inline void SetShadowMapID( GLuint id )				{ shadowMapTexID = id; }
	inline float *GetShadowMapTransposeMatrix( void )	{ return shadMapTransMatrix; }
	inline void SetShadowMapTransposeMatrix( float *m ) { memcpy( shadMapTransMatrix, m, 16*sizeof(float) ); }
	inline GLuint GetCausticMapID( void ) const			{ return causticMapTexID; }
	inline void SetCausticMapID( GLuint id ) 			{ causticMapTexID = id; }
	inline GLuint GetCausticDepthID( void ) const		{ return causticDepthMapTexID; }
	inline void SetCausticDepthID( GLuint id ) 			{ causticDepthMapTexID = id; }

	// Search paths for the scene.  This is already pretty well encapsulated,
	//    so I see no reason not to offer users direct access.
	ProgramSearchPaths *paths;
};



#endif

