// This shader combines a low resolution scattering image and a high resolution
//    image representing un-obscured (i.e., no participating media) image of the
//    scene geometry.
//
// This is roughly equivalent to the pseudocode in our ShaderX7 article, 
//    except it unrolls the loop and deals with cases where the edge-preserving
//    filter detects _NO_ valid pixels on the right side of the edge.  This
//    seems odd, but it happens frequently enough.

uniform sampler2D objColorAndDist;
uniform sampler2D lightIntegral;
uniform float extinctionCoef;

// Distance at which to pixels are considered "close enough."  This is
//   currently used as a WORLD SPACE position (not a [0..1] z-buf value).
#define DEPTH_THRESHOLD       0.5


// Compute the extiction at some distance from the eye
float ExtinctionFromDistance( float dist )
{
	return exp( - extinctionCoef * dist );
}

// Determine if two distances are close enough
bool Valid( float centerDist, float curDist )
{
	return (abs(centerDist-curDist) < DEPTH_THRESHOLD);
}

// Determine if the current distance is beyond the current pixel.
bool FarAway( float centerDist, float curDist )
{
	return ((centerDist-curDist) < -DEPTH_THRESHOLD);
}

void main( void )
{
	// We're indexing into a texture that's lower resolution than the current image.
	//    We'll be using some sort of filter over a 3x3 region, but because our texture 
	//    uses nearest-neighbor interpolation (so we can determine the actual depth for
	//    each of the "super-pixels", not linearly interpolated ones), we have to 
	//    compute the weights ourselves.  Compute the offset in the current super-pixel
	vec2 offset = gl_TexCoord[0].xy - floor(gl_TexCoord[0].xy*256)/256;
	vec2 weight = abs(offset.xy*256);

	// Load the color of the background from a previous pass (no fog is applied, so 
	//    apply a standard extinction coefficient)
	vec4 val = texture2D( objColorAndDist, gl_TexCoord[0].xy );
	val.xyz *= ExtinctionFromDistance( val.w );
	
	// Load the 3x3 pixel block of scattering data to interpolate over
	vec4 tex[9];
	tex[0] = texture2D( lightIntegral, gl_TexCoord[0].xy );
	tex[1] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( 0.0, 1.0/256.0 ) );
	tex[2] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( 0.0, -1.0/256.0 ) );
	tex[3] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( 1.0/256.0, 0.0 ) );
	tex[4] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( -1.0/256.0, 0.0 ) );
	tex[5] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( 1.0/256.0, 1.0/256.0 ) );
	tex[6] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( 1.0/256.0, -1.0/256.0 ) );
	tex[7] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( -1.0/256.0, 1.0/256.0 ) );
	tex[8] = texture2D( lightIntegral, gl_TexCoord[0].xy+vec2( -1.0/256.0, -1.0/256.0 ) );
	
	// We need to compute weights for each pixel.  We're using basically a tent filter
	//    over a 3x3 region here.  Others filters are possible, but this works pretty well.
	//    We also tried a Gaussian.  However, we can't use all of the 3x3 region, because
	//    some pixels may represent scattering for significantly longer intervals through
	//    the light.  This will cause blocky/jaggy artifacts in the scattering, which become
	//    very apparent for back-lit geometry.  So we zero a texel's weight if it's too 
	//    far from the "ideal" depth (i.e., the depth of the current pixel from the 
	//    background image)
	float weightNearby[9];
	weightNearby[0] = Valid( val.w, tex[0].w ) ? 1.0 : 0.0;
	weightNearby[1] = Valid( val.w, tex[1].w ) ? weight.y : 0.0;
	weightNearby[2] = Valid( val.w, tex[2].w ) ? 1-weight.y : 0.0;
	weightNearby[3] = Valid( val.w, tex[3].w ) ? weight.x : 0.0;
	weightNearby[4] = Valid( val.w, tex[4].w ) ? 1-weight.x : 0.0;
	weightNearby[5] = Valid( val.w, tex[5].w ) ? weight.x*weight.y : 0.0;
	weightNearby[6] = Valid( val.w, tex[6].w ) ? weight.x*(1-weight.y) : 0.0;
	weightNearby[7] = Valid( val.w, tex[7].w ) ? (1-weight.x)*weight.y : 0.0;
	weightNearby[8] = Valid( val.w, tex[8].w ) ? (1-weight.x)*(1-weight.y) : 0.0;

	// Unfortunately right at the edges (e.g., silhouettes), there might not be any 
	//    "valid" pixels according to the threshold.  This leads to black (or white)
	//    scattering color, depending on how the weights are used.  In either case it
	//    looks really odd.  What we'll do in this case is assume the ray misses the
	//    geometry, and only use scattering from adjacent texels that do not hit the
	//    nearby object (this is significantly less objectionable than using scattering
	//    along a ray that is too short, and could be thought of as "bloom" around 
	//    the edges).  Here we compute the relevant weights for this condition.
	float weightFar[9];
	weightFar[0] = FarAway( val.w, tex[0].w ) ? 1.0 : 0.0;
	weightFar[1] = FarAway( val.w, tex[1].w ) ? weight.y : 0.0;
	weightFar[2] = FarAway( val.w, tex[2].w ) ? 1-weight.y : 0.0;
	weightFar[3] = FarAway( val.w, tex[3].w ) ? weight.x : 0.0;
	weightFar[4] = FarAway( val.w, tex[4].w ) ? 1-weight.x : 0.0;
	weightFar[5] = FarAway( val.w, tex[5].w ) ? weight.x*weight.y : 0.0;
	weightFar[6] = FarAway( val.w, tex[6].w ) ? weight.x*(1-weight.y) : 0.0;
	weightFar[7] = FarAway( val.w, tex[7].w ) ? (1-weight.x)*weight.y : 0.0;
	weightFar[8] = FarAway( val.w, tex[8].w ) ? (1-weight.x)*(1-weight.y) : 0.0;
	
	// Because we've zeroed out some weights, they might not sum to what we thought.
	//    So in order to correctly normalize the color, we must sum these weights.
	float weightNearSum = weightNearby[0]+weightNearby[1]+weightNearby[2]+
	                  weightNearby[3]+weightNearby[4]+weightNearby[5]+
	                  weightNearby[6]+weightNearby[7]+weightNearby[8];
	float weightFarSum = weightFar[0]+weightFar[1]+weightFar[2]+
	                  weightFar[3]+weightFar[4]+weightFar[5]+
	                  weightFar[6]+weightFar[7]+weightFar[8];

	// Compute the weighted sums for the two conditions
	vec4 colorNear = weightNearby[0]*tex[0]+weightNearby[1]*tex[1]+weightNearby[2]*tex[2]+
	             weightNearby[3]*tex[3]+weightNearby[4]*tex[4]+weightNearby[5]*tex[5]+
	             weightNearby[6]*tex[6]+weightNearby[7]*tex[7]+weightNearby[8]*tex[8];
	vec4 colorFar = weightFar[0]*tex[0]+weightFar[1]*tex[1]+weightFar[2]*tex[2]+
	             weightFar[3]*tex[3]+weightFar[4]*tex[4]+weightFar[5]*tex[5]+
	             weightFar[6]*tex[6]+weightFar[7]*tex[7]+weightFar[8]*tex[8];
	colorNear *= (1.0/weightNearSum);
	colorFar  *= (1.0/weightFarSum);
	
	// Oddly, it's possible (in areas were there are small holes in the geometry
	//     poorly sampled by the low resolution scattering image) to have zero weight
	//     for both weightNearSum and weightFarSum.  In this case, there's no good
	//     solution (except to use a full resolution scattering image).  We default
	//     to using the scattering for the single nearest texel.  This seems to work
	//     pretty well.
	vec4 color = (weightFarSum > 0.05) ? colorFar : tex[0];
	
	// Ok.  Combine the final colors together
	gl_FragColor.xyz = (weightNearSum > 0.05 ? colorNear : color) + val.xyz; 
}

