// This is the shader that implements the basic version of our hybrid
//   volumetric shadow algorithm, which generates volumetric shadows
//   from a single-colored point light in homogeneous participating media.

#version 120
#extension GL_EXT_gpu_shader4 : enable

// The number of samples to use when ray marching.  Artifacts?  Increase this.
#define NUM_FOG_SAMPLES 50

// Textures needed by the shader
uniform sampler2D objColorAndDist;  // Color of (RGB) & distance to (A) surfaces in scene
uniform sampler2DShadow shadowMap;  // Scene shadow map
uniform sampler2DArray shadowVolumeDists;  // Distances to formost and backmost shadow volumes
uniform sampler2D newFogTex;        // The reparameterized F texture described in the text
uniform float lightIntensity;       // Modifier to light intensity 
uniform float extinctionCoef;       // Fog density

// Currently set (in the C++ code) to 1 all the time.  We were using this to 
//    swap back and forth between two fog models.  Do a search for "Sun et al", and 
//    uncomment relevant lines to use the Sun et al model.  Please beware that the Sun et al. 
//    model and the Hoffman and Preetham models require _different_ light intensities and 
//    extinction coefficients to achieve a similar result.
uniform float useAlternateFog;  

// Used for shadow mapping (eye-space S,T,R, & Q planes)
uniform vec4 epS, epT, epR, epQ;

// Values to invert projection & viewport matrix to allow us to compute eye-space direction
//   given a screen space coordinate
uniform vec2 imgSize;
uniform float tan_fovy_2, farDist;


// Compute the extiction at some distance from the eye
float ExtinctionFromDistance( float dist )
{
	return exp( - extinctionCoef * dist );
}

// Code to use the Hoffman and Preetham model using very specific, hardcoded values 
//    that we found useful.  We've modified this to take in parameters nearly identical 
//    to our reparameterized F-texture for easy swapping between fog models.
float AlternateFog( float cosVal, float distToFrag )
{
	float fex = exp( -distToFrag*(0.15+0.15) );
	float ins = (1-fex);
	float brt = 3.0/(16.0*3.14159)*0.15*(1.0+cosVal*cosVal);
	float bmt = 1.0/(4.0*3.14159)*0.15*(1.0-0.95)*(1.0-0.95)/(pow((1.0+0.95*0.95-2*0.95*cosVal),1.5) );
	float lint = (brt+bmt)/(0.15+0.15)*ins;
	return lint;
}

// Turns out much of the Hoffman & Preetham model can be factored out into a constant
//    term, which also helps eliminate numerical precision issues when accumulating
//    scattering over small intervals.  The following two functions split the model into 2
float AlternateFogDistanceTerm( float distToFrag )
{
	float fex = exp( -distToFrag*(0.15+0.15) );
	return (1-fex);
}
float AlternateFogConstant( float cosVal )
{
	float brt = 3.0/(16.0*3.14159)*0.15*(1.0+cosVal*cosVal);
	float bmt = 1.0/(4.0*3.14159)*0.15*(1.0-0.95)*(1.0-0.95)/(pow((1.0+0.95*0.95-2*0.95*cosVal),1.5) );
	return (brt+bmt)/(0.15+0.15);
}

// Check if a point is illuminated by checking the shadow map.
float IsPointIlluminated( vec4 eyeSpacePosition )
{
	vec4 smapCoord;
	smapCoord.x = dot( epS, eyeSpacePosition );
	smapCoord.y = dot( epT, eyeSpacePosition );
	smapCoord.z = dot( epR, eyeSpacePosition );
	smapCoord.w = dot( epQ, eyeSpacePosition );
	smapCoord /= smapCoord.w;
	return ( all(equal(smapCoord.xyz,clamp(smapCoord.xyz,vec3(0),vec3(1)))) ? 
					shadow2D( shadowMap, smapCoord.xyz ).z : 
					1 );
}

// Takes a 2D location in screen space and returns an eye-space direction
//    corresponding to this pixel.  Using this function reduces the number
//    of values we need to pass down.  Instead of texture with 3D positions
//    of geometry in the scene, we just need a depth image and we can 
//    compute positions from their depth and screen-space position.
vec3 DirectionFromPos( vec2 screenPos )
{
	vec3 dir;
	dir.xy = tan_fovy_2 * (( 2*screenPos - imgSize ) / imgSize);
	dir.x *= imgSize.x / imgSize.y;
	dir.z = -1;
	return normalize( dir );
}

void main( void )
{

	// Find the relevent positions and colors
	vec4 fragColorAndPos = texture2D( objColorAndDist, gl_TexCoord[0].xy );	  // Pos & color of scene geom
	vec4 frontVolPos = texture2DArray( shadowVolumeDists, vec3( gl_TexCoord[0].xy, 0 ) );  // Front shadow vol
	vec4 backVolPos = texture2DArray( shadowVolumeDists, vec3( gl_TexCoord[0].xy, 1 ) );   // Back shadow vol
	
	// Compute some distances (really just renaming them here...  other experimental shaders had to do computations)
	float distToFrag  = fragColorAndPos.w; 
	float distToFrontVol = frontVolPos.z ;
	float distToBackVol = backVolPos.z;   
	
	// However, the distance to front volume & back volume ARE NOT WELL DEFINED, since not all pixels
	//    hit these volumes (plus these volumes could be behind scene geometry, etc).  Thus we check to
	//    make sure our front-to-back interval is valid and as small as possible.
	distToBackVol = (distToFrag < 0.01 ? distToFrag : distToBackVol);
	distToBackVol = (distToFrag < 0.01 || distToBackVol < 0.01 || distToBackVol+0.1 > distToFrag) ? distToFrag : distToBackVol+0.1;

	// Compute in-scattering using the fog model by doing some ray marching.
	//    First, however, compute some values common throughout the loop as well as
	//    the contribution for the first illuminated interval (i.e., from eye to front shadow volume)
	vec3 eyeToLight = normalize( gl_LightSource[0].position.xyz );
	vec3 eyeToFrag = DirectionFromPos( gl_FragCoord.xy ); 
	float cosGamma = dot( eyeToLight, eyeToFrag );
	float cosGammaLookup = 0.5*cosGamma+0.5;
	
	// A constant value we factor out to eliminate precision issues when accumulating scattering
	float fogConstant = AlternateFogConstant( cosGamma );   // For Hoffman and Preetham
	//float fogConstant = 1;                                // For Sun et al Model      
	
	// Fog contribution at the eye and the surface of the scene geometry 
	float fogContrib_frag = AlternateFogDistanceTerm( distToFrag );                                    // For Hoffman and Preetham
	//float fogContrib_frag = texture2D( newFogTex, vec2( cosGammaLookup, distToFrag/50 ) ).x;         // For Sun et al model
	float fogContrib_eye = AlternateFogDistanceTerm( 0 );                                              // For Hoffman and Preetham
	//float fogContrib_eye = texture2D( newFogTex, vec2( cosGammaLookup, 0.0 ) ).x;                    // For Sun et al model
	
	// If we don't hit the surface of a shadow volume, we can use a simple computation
	//   to determine scattering over the entire interval, all at once.  If so, we're done.
	if ( distToFrontVol < 0.01 || distToFrontVol > distToFrag)
	{
		gl_FragColor.xyz = vec3( lightIntensity * fogConstant * ( fogContrib_frag - fogContrib_eye ) );
		gl_FragColor.w   = distToFrag;
		return;
	}
	
	
	// OK, we did hit a shadow volume.  Compute scattering at this interface.
	float fogContrib_frontVol  = AlternateFogDistanceTerm( distToFrontVol );                             // For Hoffman and Preetham
	//float fogContrib_frontVol = texture2D( newFogTex, vec2( cosGammaLookup, distToFrontVol/50.0 ) ).x; // For Sun et al model
	
	
	// Now we're going to march from front shadow volume to back volume
	float lastFogContrib = fogContrib_frontVol;
	float distDelta = ( distToBackVol - distToFrontVol ) / NUM_FOG_SAMPLES;
	float curDist = distToFrontVol;
	float inScattering = 0;
	float lit = 0.0;
	for ( int i = 1; i < NUM_FOG_SAMPLES; i++ )
	{
		// We've moved along the ray one step.
		curDist += distDelta;
	
		// Check if our new position is illuminated
		lit = IsPointIlluminated( vec4( curDist * eyeToFrag, 1 ) );
			
		// We either entered or left a shadowed region.  We'll need to know the fog 
		//    contribution at this point, so grab it.
		float currFogContrib = AlternateFogDistanceTerm( curDist );                             // For Hoffman and Preetham
		//float currFogContrib = texture2D( newFogTex, vec2( cosGammaLookup, curDist/50 ) ).x;  // For Sun et al model
		
		// However, if we're currently shadowed, then the last segment is inside shadow,
		//    so we don't want to count its fog contribution. 
		inScattering += lit*(currFogContrib - lastFogContrib);
		
		// Remember this contribution for computations over the next interval (so we need
		//    not recompute by looking into the texture again).  
		lastFogContrib = currFogContrib;	
		
		// Moved along the ray one more step.
		//curDist += distDelta;	
	}
				
	// Compute the contribution of light just outside the shadow, as we need to use that to compute
	//    scattering between there and where we finally hit the scene geometry.
	curDist = distToBackVol;
	float finalContrib = AlternateFogDistanceTerm( curDist );                             // For Hoffman and Preetham
	//float finalContrib = texture2D( newFogTex, vec2( cosGammaLookup, curDist/50 ) ).x;  // For Sun et al model
	
	// Oddly, we occasionally have some numerical errors on really thin shadows where we
	//   end up showing MORE scattering while travelling through shadow than if we didn't
	//   encounter a shadow at all!  Clearly this is wrong, so make sure we get rid of this
	//   by clamping to the unshadowed contribution.
	inScattering = min( finalContrib - fogContrib_frontVol, inScattering );
	
	// Now, add to the scattering between shadow volumes the contributions from the first interval
	//   (i.e., eye to front shadow volume) and the last interval (back shadow volume to scene geometry)
	inScattering = fogConstant*(fogContrib_frontVol - fogContrib_eye + fogContrib_frag - finalContrib + inScattering);
	
	// Finally, add modulate this scattering by our arbitrary "light intensity" factor.
	gl_FragColor.xyz = vec3(lightIntensity * inScattering);
	gl_FragColor.w   = distToFrag;
}