// This is the shader that implements the more complex version of our hybrid
//   volumetric shadow algorithm, which generates volumetric shadows
//   from a textured spot light in homogeneous participating media.

#version 120
#extension GL_EXT_gpu_shader4 : enable

// To make this code simpler, we simplified the step sizes.  We perform
//    3 loops (one before the shadow, one between the shadow volumes, 
//    and one after the shadow).  Each one can use up to MAX_NUM_SAMPLES
//    step.  Each step is either OUTSIDE_DELTA world-space units for the
//    first or last loops or INSIDE_DELTA world-space units for the 
//    inner (shadow) loop.  To reduce unnecessary samples, a more
//    sophisticated sampling approach, such as the one described in the 
//    text, should be used.
#define MAX_NUM_SAMPLES  100
#define OUTSIDE_DELTA    0.1
#define INSIDE_DELTA     0.03


// Currently set (in the C++ code) to 1 all the time.  We were using this to 
//    swap back and forth between two fog models.  While we kept the simple
//    hybrid easy to swap back and forth to the Sun et al model, we removed
//    that code from this shader to keep it as legible as possible.  See the
//    simpler shader to see how the Sun et al model is used, and it should be
//    relatively easy to implement here.  Note:  We weren't real pleased by the 
//    results of the Sun et al model under texture spotlights (which is another
//    reason we removed it from this shader)
uniform float useAlternateFog;

// Textures needed by the shader
uniform sampler2D objColorAndDist;         // Color of (RGB) & distance to (A) surfaces in scene
uniform sampler2D lightVolBounds;          // Distances to front & back of light frustum
uniform sampler2D lightTexColor;           // Spotlight color texture
uniform sampler2DShadow shadowMap;         // Scene shadow map
uniform sampler2D newFogTex;               // The reparameterized F texture described in the text
uniform sampler2DArray shadowVolumeDists;  // Distances to formost and backmost shadow volumes

// Used for shadow mapping (eye-space S,T,R, & Q planes)
uniform vec4 epS, epT, epR, epQ;

// Some scene parameters
uniform float lightIntensity;              // Modifier to light intensity 
uniform float extinctionCoef;              // Fog density

// Values to invert projection & viewport matrix to allow us to compute eye-space direction
//   given a screen space coordinate
uniform vec2 imgSize;
uniform float tan_fovy_2, farDist, sin_fovy_2;


// Compute the extiction at some distance from the eye
float ExtinctionFromDistance( float dist )
{
	return exp( - extinctionCoef * dist );
}


// Code to use the Hoffman and Preetham model using very specific, hardcoded values 
//    that we found useful.  We've modified this to take in parameters nearly identical 
//    to our reparameterized F-texture for easy swapping between fog models.
float AlternateFog( float cosVal, float distToFrag )
{
	float fex = exp( -distToFrag*(0.15+0.15) );
	float ins = (1-fex);
	float brt = 3.0/(16.0*3.14159)*0.15*(1.0+cosVal*cosVal);
	float bmt = 1.0/(4.0*3.14159)*0.15*(1.0-0.95)*(1.0-0.95)/(pow((1.0+0.95*0.95-2*0.95*cosVal),1.5) );
	float lint = (brt+bmt)/(0.15+0.15)*ins;
	return lint;
}

// Check if a point is illuminated by checking the shadow map.  If it is illuminated,
//    return the correct color, using the spotlight texture.
vec4 IsPointIlluminated( vec4 eyeSpacePosition )
{
	vec4 smapCoord;
	smapCoord.x = dot( epS, eyeSpacePosition );
	smapCoord.y = dot( epT, eyeSpacePosition );
	smapCoord.z = dot( epR, eyeSpacePosition );
	smapCoord.w = dot( epQ, eyeSpacePosition );
	smapCoord /= smapCoord.w;
	vec4 lightColor = texture2DLod( lightTexColor, smapCoord.xy, 0.0 );
	smapCoord.z = smapCoord.z <= 0 ? 0 : smapCoord.z;
	smapCoord.z = (length(eyeSpacePosition.xyz-gl_LightSource[0].position.xyz) < 0.5 ? -1 : smapCoord.z);
	return ( all(equal(smapCoord.xyz,clamp(smapCoord.xyz,vec3(0.02,0.02,0),vec3(0.98)))) ? 
					lightColor*shadow2D( shadowMap, smapCoord.xyz ) : vec4(0) );
}

// Merely checks if a point is lit (using the shadow map), without getting a color from
//    the spotlight texture
bool IsLit( vec4 eyeSpacePosition )
{
	vec4 smapCoord;
	smapCoord.x = dot( epS, eyeSpacePosition );
	smapCoord.y = dot( epT, eyeSpacePosition );
	smapCoord.z = dot( epR, eyeSpacePosition );
	smapCoord.w = dot( epQ, eyeSpacePosition );
	smapCoord /= smapCoord.w;
	return ( all(equal(smapCoord.xyz,clamp(smapCoord.xyz,vec3(0.02,0.02,0),vec3(0.98)))) ? 
					shadow2D( shadowMap, smapCoord.xyz ).x > 0 : true );
}

// Simply gets a color from the spotlight texture without checking the shadow map for visibility
vec4 GetColor( vec4 eyeSpacePosition )
{
	vec3 smapCoord;
	smapCoord.x = dot( epS, eyeSpacePosition );
	smapCoord.y = dot( epT, eyeSpacePosition );
	smapCoord.z = dot( epQ, eyeSpacePosition );
	smapCoord /= smapCoord.z;
	return texture2DLod( lightTexColor, smapCoord.xy, 0.0 );
}

// Takes a 2D location in screen space and returns an eye-space direction
//    corresponding to this pixel.  Using this function reduces the number
//    of values we need to pass down.  Instead of texture with 3D positions
//    of geometry in the scene, we just need a depth image and we can 
//    compute positions from their depth and screen-space position.
vec3 DirectionFromPos( vec2 screenPos )
{
	vec3 dir;
	dir.xy = tan_fovy_2 * (( 2*screenPos - imgSize ) / imgSize);
	dir.x *= imgSize.x / imgSize.y;
	dir.z = -1;
	return normalize( dir );
}

void main( void )
{
	// Find the region of the ray that is potentially illuminated
	vec4 lightBounds = texture2D( lightVolBounds, gl_TexCoord[0].xy );
	
	// Check if we even cross the light volume...
	if (lightBounds.z+lightBounds.x < 0.5) 
		discard;
		
	// Find the relevent positions
	vec4 fragPos = texture2D( objColorAndDist, gl_TexCoord[0].xy );	
	float distToFrag = fragPos.w; 
	vec3 eyeToFrag = DirectionFromPos( gl_FragCoord.xy ); 
	
	// Check if the eye point is lit (0.1 is the near plane).
	bool eyeLit = IsLit( vec4( 0.1*eyeToFrag, 1 ) );
	
	// Find out where we intersect the shadow volume (if at all)
	float frontShadowVol = texture2DArray( shadowVolumeDists, vec3( gl_TexCoord[0].xy, 0 ) ).z;
	float backShadowVol = texture2DArray( shadowVolumeDists, vec3( gl_TexCoord[0].xy, 1 ) ).z;
	bool hitShadowVol = (frontShadowVol > 0 && frontShadowVol < fragPos.w);
	
	// Compute how far it is to the light
	float distToLight = length( gl_LightSource[0].position.xyz );
	
	// Now determine where we'll start marching, and how far we'll march	
	float startDist = (lightBounds.z > 0.5 ? lightBounds.w : 0 );
	float distLightToStartPos = length( gl_LightSource[0].position.xyz - startDist*eyeToFrag );
	float endDist = (lightBounds.x > 0.5 ? min(distToFrag,lightBounds.y) : distToFrag);
	float firstLoopEndDist = ( hitShadowVol ? frontShadowVol : endDist );
	float secondLoopEndDist = ( hitShadowVol && backShadowVol > 0 ? min(backShadowVol,endDist) : endDist );
	
	// Data for the airlight in-scattering using the airlight model 	
	float cosGamma = dot( normalize( gl_LightSource[0].position.xyz ), eyeToFrag );
	float lastFogContrib = AlternateFog( cosGamma, startDist );
	
	// Initialize the data for the ray marching.
	vec4 lightColor = vec4(0);
	
	// Factors used to speed up the loop.  We use an r^2 fall off for the colored texture
	//    otherwise it looks _really_ odd.  These factors are needed for that.
	float factor1 = distToLight*distToLight;
	float factor2 = 2*distToLight*cosGamma;
	
	// Loop from the front of the light volume to the front shadow volume.  Notice
	//    that we know we're lit here, so we can simply look into the spotlight texture
	for ( int i = 1; i < MAX_NUM_SAMPLES; i++ )
	{
		float curDist = min( firstLoopEndDist, startDist+i*OUTSIDE_DELTA );		
		float currFogContrib = AlternateFog( cosGamma, curDist );
		vec4 color = GetColor( vec4( curDist * eyeToFrag, 1 ) );
	    color = color*4/(factor1+curDist*curDist-curDist*factor2);
		lightColor +=  color * (currFogContrib - lastFogContrib);	
		lastFogContrib = currFogContrib;
		if (curDist >= firstLoopEndDist) break;
	}
	
	// If we didn't encounter a shadow volume, we've accumulated the correct color.  Return.
	gl_FragData[0] = lightIntensity * lightColor * 50;  // 50 is a factor to brighten it up.
	gl_FragData[0].w = fragPos.w;                       //     as the r^2 falloff dims things
	gl_FragData[1] = lightIntensity * lightColor * 50;  //     significantly.
	gl_FragData[1].w = fragPos.w;
	if (!hitShadowVol) return;
	
	// OK.  We hit a shadow volume.  March inside the shadow volumes, looking to see
	//    if we're lit and grabbing the light color if so.
	for ( int i = 1; i < MAX_NUM_SAMPLES; i++ )
	{
		float curDist = min( secondLoopEndDist, firstLoopEndDist+i*INSIDE_DELTA );		
		float currFogContrib  = AlternateFog( cosGamma, curDist );
		vec4 color = IsPointIlluminated( vec4( curDist * eyeToFrag, 1 ) );
	    color = color*4/(factor1+curDist*curDist-curDist*factor2);
		lightColor +=  color * (currFogContrib - lastFogContrib);	
		lastFogContrib = currFogContrib;
		if (curDist >= secondLoopEndDist) break;
	}
	
	// If our scene surface is in shadow, there's no need to perform the third loop.
	gl_FragData[0] = lightIntensity * lightColor * 50; 
	gl_FragData[0].w = fragPos.w;
	gl_FragData[1] = lightIntensity * lightColor * 50;
	gl_FragData[1].w = fragPos.w;
	if (secondLoopEndDist >= endDist) return;
	
	// Alright, march from the end of the shadow volume until we leave the light 
	//    frustum or encounter the scene geometry.  
	for ( int i = 1; i < MAX_NUM_SAMPLES; i++ )
	{
		float curDist = min( endDist, secondLoopEndDist+i*OUTSIDE_DELTA );		
		float currFogContrib  = AlternateFog( cosGamma, curDist );
		vec4 color = GetColor( vec4( curDist * eyeToFrag, 1 ) );
	    color = color*4/(factor1+curDist*curDist-curDist*factor2);
		lightColor +=  color * (currFogContrib - lastFogContrib);	
		if (curDist >= endDist) break;
		lastFogContrib = currFogContrib;
	}
	
	// OK!  We're done!
	gl_FragData[0] = lightIntensity * lightColor * 50; 
	gl_FragData[0].w = fragPos.w;
	gl_FragData[1] = lightIntensity * lightColor * 50;
	gl_FragData[1].w = fragPos.w;
}

