/***************************************************************************/
/* sceneLoader.cpp                                                         */
/* -----------------------                                                 */
/*                                                                         */
/* This program is a common framework I use for most of my research.  It   */
/*    reads a custom-designed scene file from a file into a global         */
/*    scenegraph class (Scene), which encapsulates a lot of functionality  */
/*    I commonly use in my work.  It, more or less, allows me to change    */
/*    just the code inside the "RenderingTechniques/" directory as I move  */
/*    from project to project.                                             */
/*                                                                         */
/* Most of the relevant code for this demo is in this file, the directory  */
/*    "RenderingTechniques", and the shaders in "bin/shaders".             */
/*                                                                         */
/* Chris Wyman (07/22/2008)                                                */
/***************************************************************************/

#include <string.h>
#include "sceneLoader.h"
#include "RenderingTechniques/renderingData.h"
#include "RenderingTechniques/renderingMethods.h"


// I'm a C programmer at heart.  To reduce the ugliness of my global 
//    variables, they are organized into two relatively nice structures, the
//    Scene class, which stores just about everything related to the scene 
//    data, and theRenderingData class, which stores everything else (GL 
//    state, UI data)
Scene *scene = 0;
extern RenderingData *data;


// This is the GLUT callback for drawing/refreshing the screen image.
void DisplayCallback( void )
{
	// Code to start timing the frame.  
	data->ui->frameSpeed->StartFrame();

	// Check if we have some special rendering mode (like volumetric shadows)
	if (*data->ui->displayID > 0)
	{
		switch( *data->ui->displayID )
		{
		case 1:
			Display_WithVolumetricShadows();
			break;
		case 2:
			Display_WithBasicVolumetricShadows();
			break;
		}
	}

	// Nope.  Let's use a default, "simple" OpenGL rendering.
	else
	{
		// Create a shadow map
		scene->CreateShadowMap( data->fbo->shadowMap, 
			                    data->param->shadowMatrixTrans, 
								0, // Light Number
								*(data->param->shadowMapBias) );

		// The light intensity is a modifiable value, but it mainly makes this
		//     "simple" OpenGL rendering way too bright or way too dark.  To 
		//     account for the fact this modifier must be cranked way up for volumetric
		//     shadows to get an acceptable brightness, I'll turn it off for this mode.
		float oldLightIntensity = scene->GetLightIntensityModifier();
		scene->SetLightIntensityModifier( 1.0 );

		// Draw the scene simply.
		data->fbo->mainWin->BindBuffer();
		data->fbo->mainWin->ClearBuffers();
			glLoadIdentity();
			scene->LookAtMatrix();
			scene->SetupEnabledLightsWithCurrentModelview();
			scene->Draw( MATL_FLAGS_USESHADOWMAP );
		data->fbo->mainWin->UnbindBuffer();

		// Set the light intensity back to "normal"
		scene->SetLightIntensityModifier( oldLightIntensity );
	}

	// Copy the final rendering to the screen.  This "extra frame copy" is a bit silly 
	//    and drags down performance most of the time, but is really useful for rendering
	//    at much larger resolutions than can physically be displayed, either to downsample
	//    for anti-aliasing or for generating larger images for publication.  In my research
	//    code, this small extra overhead is worth it.
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	data->fbo->mainWin->DisplayAsFullScreenTexture( FBO_COLOR0 );

	// If a screen capture is desired, do it now (before writing timer text)
	if (data->ui->captureScreen)
	{
		data->ui->frameGrab->CaptureFrame();
		data->ui->captureScreen = false;
	}

	// If the user needs help, display the UI help screen
	if (data->ui->displayHelp)
		DisplayHelpScreen( data->fbo->helpScreen );


	// Display the framerate and finish frame
	DisplayTimer( data->ui->frameSpeed->EndFrame(), 1024, 1024 );
	glFlush();
	glutSwapBuffers();
}


void PrintDemoData( void )
{
	printf("**************************************************************************\n");
	printf("*       Interactive Volumetric Shadows in Participating Media with       *\n");
	printf("*                         Single-Scattering                              *\n");
	printf("*     (Proceedings of the IEEE Symposium on Interactive Ray Tracing)     *\n");
	printf("*                Chris Wyman               Shaun Ramsey                  *\n");
	printf("**************************************************************************\n");
}

void CheckMachineCapabilities( void )
{
	printf("    (-) Running: OpenGL v%s, GLEW v%s, GLSL v%s \n", 
		glGetString( GL_VERSION ), 
		glewGetString( GLEW_VERSION ),
		glGetString(GL_SHADING_LANGUAGE_VERSION) );
	printf("    (-) Compiled against: GLUT v3.7.6 \n");
	printf("    (-) Running on a: %s %s\n", glGetString( GL_VENDOR ), glGetString( GL_RENDERER ));
	glGetError();
	
	/* check if the machine is equipped to run the demo */
	if (!GLEW_VERSION_2_0)
	{
		printf("    (-) FATAL ERROR: Demo requires OpenGL 2.0 or higher!\n");
		exit(-1);
	}
	if (!GLEW_EXT_gpu_shader4)
	{
		printf("    (-) FATAL ERROR: Demo requires OpenGL Extension EXT_gpu_shader4!\n");
		exit(-1);
	}
	if (!GLEW_EXT_geometry_shader4)
	{
		printf("    (-) FATAL ERROR: Demo requires OpenGL Extension EXT_geometry_shader4!\n");
		exit(-1);
	}
	if (!GLEW_EXT_texture_array)
	{
		printf("    (-) FATAL ERROR: Demo requires OpenGL Extension EXT_texture_array!\n");
		exit(-1);
	}
}



int main(int argc, char* argv[] )
{
	// Program setup & command line parsing
	bool verbose = false;
	char scenefile[ 256 ] = "";
	PrintDemoData();

	printf("(+) Parsing any command line parameters...\n");
	if (argc < 1)
	{
		printf("Usage: %s <sceneFile>\n", argv[0]);
		exit(0);
	}
	for( int i=1; i < argc; i++ )
	{
		if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help") || 
			!strcmp(argv[i], "-?") || !strcmp(argv[i], "/?") )
		{
			printf("Usage: %s <sceneFile>\n", argv[0]);
			exit(0);
		}
		else if (!strcmp(argv[i], "-v") || !strcmp(argv[i], "-verbose"))
			verbose = true;
		else
		{
			strncpy( scenefile, argv[i], 255 );
			break;
		}
	}

	// Load the scene file.  If none was specified, use the default.
	if (!strcmp(scenefile, ""))
	{
		char *ptr, *token=argv[0];
		ptr = strtok( argv[0], "/\\" );
		while (ptr)  // Why does MS not have strrtok?
		{
			token = ptr;
			ptr = strtok( NULL, "/\\" );
		}
		printf("(+) No scene specified on the command line, using default 'boxScene.txt'\n");
		printf("    (-) Specify by running as \"%s <sceneFile>\".\n", token);
		strcpy( scenefile, "boxScene.txt" );
	}
	scene = new Scene( scenefile, verbose );

	// Start the GL window
	printf("(+) Initializing OpenGL state...\n");
	glutInit(&argc, argv);
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA );
	glutInitWindowSize( scene->GetWidth(), scene->GetHeight() );
	glutCreateWindow( "Demo: Interactive Volumetric Shadows in Participating Media with Single Scattering" );
	glewInit();

	// Now that GLEW has initialized, double check that we can actually run this demo.
	CheckMachineCapabilities();

	// Set the GLUT callback functions 
	glutDisplayFunc( DisplayCallback );
	glutReshapeFunc( ReshapeCallback );
	glutIdleFunc( IdleCallback );
	glutMouseFunc( MouseButtonCallback );
	glutMotionFunc( MouseMotionCallback );
	glutKeyboardFunc( KeyboardCallback );
	glutSpecialFunc( SpecialKeyboardCallback ); 

	// Make sure any preprocessing that needs to be done occurs
	scene->Preprocess();

	// Initialized various data needed in the various rendering (display) callbacks
	InitializeRenderingData();

	// Run the display!
	printf("(+) Starting Main Loop...\n" );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_LIGHTING );
	glEnable( GL_NORMALIZE );
	glEnable( GL_CULL_FACE );
	scene->GetLight(0)->Enable();
	glutMainLoop();

	return 0;
}
